class Director extends GameObject {
  protected float gameTime = 0;
  protected float nextWaveTime = 0;
  float waveInterval = 20;
  float waveRating = 70;
  int waveNumber = 0;
  int maxEnemies = 40;
  int maxShipsPerWave = 30;
  
  float linearGrowth = 10;
  float exponentialGrowth = 1.01;
  
  protected List<ShipTemplate> templates;
  
  Director(){
    super();
	drawOnUI = true;
    
    templates = new ArrayList<ShipTemplate>();
	addTemplate("lancer", new LancerFactory(), 1.5, 4.0, 20, 50, 90000, 50, 10.0, 7);
	addTemplate("kamikaze", new KamikazeFactory(), 1.0, 2.0, 7, 10, 70000, 70, 30.0, 5);
    addTemplate("mothership", new MothershipFactory(), 5.0, 7.0, 500, 500, 900000, 30, 50.0, 2);
  }
  
  void update(float dt){
	//Pause the waves if there are too many enemies already on screen
	if ( enemiesStillAlive > maxEnemies ){
		return;
	}
	
	if ( enemiesStillAlive > 0 ){
	  gameTime += dt;
	} else {
	  gameTime += dt * 2.5;
	}

    if (nextWaveTime <= gameTime){
      spawnWave(waveRating);
      nextWaveTime = gameTime + waveInterval;
      waveRating = (waveRating + linearGrowth) * exponentialGrowth;
    }    
  }
  
  void spawnWave(float rating){
    waveNumber++;
    int spawnedShips = 0;
	
    float availableRating = rating;
    while ( (availableRating > 0) && (spawnedShips < maxShipsPerWave) ){
      ShipTemplate template = pickRandomTemplate(availableRating);
      if ( template == null ){
        break;
      }

      int shipCount = min(floor(availableRating / template.averageRating), template.maxShips);
      Vec2 regionCorner = new Vec2(
        random(template.maxRadius + 1, worldHalfWidth - template.maxRadius - 1) * randomSign(), 
        random(template.maxRadius + 1, worldHalfHeight - template.maxRadius - 1) * randomSign()
      );
      float regionWidth = random(template.maxRadius * sqrt(shipCount) * 2, template.maxRadius * shipCount * 2);
      float regionHeight = random(template.maxRadius * sqrt(shipCount) * 2, template.maxRadius * shipCount * 2);
	  float spawnDelay = random(0, waveInterval/2);
	  float spawnDelayVariance = 1.0;
	  float spawnDuration = 3;
      
      for(int i = 0; i < shipCount; i++){
        //Randomize ship properties based on the template
        float radius = random(template.minRadius, template.maxRadius);
        float maxHitpoints = map(radius, template.minRadius, template.maxRadius, template.minHitpoints, template.maxHitpoints);
		
        //Place the ship somewhere inside the selected region
        Vec2 position = new Vec2(random(0, regionWidth), random(0, regionHeight));
        position.addLocal(regionCorner);
        position.x = constrain(position.x, -worldHalfWidth + radius, worldHalfWidth - radius);
        position.y = constrain(position.y, -worldHalfHeight + radius, worldHalfHeight - radius);
        
        Ship ship = template.factory.newShip(position, radius, maxHitpoints);
        ship.maxThrust = template.thrust;
        ship.maxSpeed = template.speed;
        
        entities.add( new ShipSpawner(ship, spawnDelay + random(0, spawnDelayVariance), spawnDuration) );
		
		spawnedShips++;
		if ( spawnedShips > maxShipsPerWave ){
			break;
		}
      }
      
      availableRating = availableRating - shipCount * template.averageRating;
    }
  }
  
  void addTemplate(
    String debugName,
    ShipFactory factory,
    float minRadius, float maxRadius,
    float minHitpoints, float maxHitpoints,
    float thrust,
    float speed,
    float averageRating,
    int maxShips
  ){
    ShipTemplate template = new ShipTemplate(debugName, factory, minRadius, maxRadius, minHitpoints, maxHitpoints, thrust, speed, averageRating, maxShips);
    templates.add(template);
  }
  
  ShipTemplate pickRandomTemplate(float maxRating){
    List<ShipTemplate> usableTemplates = new ArrayList<ShipTemplate>();
    for(ShipTemplate template : templates){
      if (template.averageRating <= maxRating){
        usableTemplates.add(template);
      }
    }
    if ( usableTemplates.size() > 0 ) {
      return usableTemplates.get( int(random(0, usableTemplates.size())) );
    } else {
      return null;
    }    
  }
  
  void drawUI(){
    pushMatrix();
      rectMode(CORNER);
      int border = 10;
	  
      float waveWait = constrain((nextWaveTime - gameTime)/waveInterval, 0, 1);
      float waveIndicatorWidth = 100, waveIndicatorHeight = 10;
      translate(width - waveIndicatorWidth - border, border);
	  
      rectMode(CORNER);
      stroke(0);
      fill(255);
      rect(0, 20, waveIndicatorWidth, waveIndicatorHeight);
      noStroke();
      fill(50, 50, 50);
      rect(0, 20, waveIndicatorWidth * (waveWait), waveIndicatorHeight);
	  
	  textFont(font);
  	  textAlign(LEFT, TOP);
	  textSize(20);
	  fill(0);
      text("WAVE: " + waveNumber, 0, 0);
      
    popMatrix();
  }
}

class ShipTemplate {
  String debugName;
  ShipFactory factory;
  float minRadius, maxRadius;
  float minHitpoints, maxHitpoints;
  float thrust;
  float speed;
  float averageRating;
  int maxShips;
  
  ShipTemplate(
    String debugName,
    ShipFactory factory,
    float minRadius, float maxRadius,
    float minHitpoints, float maxHitpoints,
    float thrust,
    float speed,
    float averageRating,
    int maxShips
  ){
    this.debugName = debugName;
    this.factory = factory;
    this.minRadius = minRadius;
    this.maxRadius = maxRadius;
    this.minHitpoints = minHitpoints;
    this.maxHitpoints = maxHitpoints;
    this.thrust = thrust;
    this.speed = speed;
    this.averageRating = averageRating;
    this.maxShips = maxShips;
  }
}

float randomSign(){
  if ( random(0.0, 1.0) > 0.5 ){
    return 1;
  } else {
    return -1;
  }
}

class ShipSpawner extends GameObject {
  Ship shipToSpawn;
  float spawnDelay, spawnDuration, elapsedTime = 0;
  
  ShipSpawner(Ship shipToSpawn, float spawnDelay, float spawnDuration){
    this.shipToSpawn = shipToSpawn;
    this.spawnDelay = spawnDelay;
    this.spawnDuration = spawnDuration;
  }
  
  void update(float dt){
    if ( dead ){
      return;
    }
    
    elapsedTime += dt;
    if ( elapsedTime >= spawnDelay + spawnDuration ){
      spawnTheShip();
      die();
    }
  }
  
  void draw(){
    if ( elapsedTime < spawnDelay ){
      return;
    }
    pushMatrix();
      translate(shipToSpawn.startPosition.x, shipToSpawn.startPosition.y);
      float phase = constrain((elapsedTime - spawnDelay) / spawnDuration, 0.0, 1.0);
      fill( shipToSpawn.baseColor, map(phase, 0, 1, 80, 250));
      noStroke();
      ellipse(0, 0, shipToSpawn.radius*2*phase*1.05, shipToSpawn.radius*2*phase*1.05);
    popMatrix();
  }
  
  void spawnTheShip(){
    entities.add(shipToSpawn);
  }
}

class ShipFactory{
  Ship newShip(Vec2 position, float radius, float hitpoints){
    return null;
  }
}

class MothershipFactory extends ShipFactory{
  Mothership newShip(Vec2 position, float radius, float hitpoints){
    return new Mothership(position, radius, hitpoints);
  }
}

class LancerFactory extends ShipFactory{
  Lancer newShip(Vec2 position, float radius, float hitpoints){
    return new Lancer(position, radius, hitpoints);
  }
}

class KamikazeFactory extends ShipFactory{
  Kamikaze newShip(Vec2 position, float radius, float hitpoints){
    return new Kamikaze(position, radius, hitpoints);
  }
}
