class Particle {
	Vec2 position, velocity, acceleration;
	float damping = 0.0;
	float age, lifetime;
	color startColor = color(100, 100, 100, 200);
	color endColor = color(100, 100, 100, 0);
	float size = 1;
	boolean dead = false;
	
	Particle(){
		//...stub
	}
	
	void update(float dt){
		position.addLocal(velocity.mul(dt));
		velocity.addLocal(acceleration.mul(dt));
		Vec2 dampingAcc = velocity.mul(-damping * dt);
		velocity.addLocal(dampingAcc);
		
		age += dt;
		if ( age > lifetime ){
			dead = true;
		}
	}
	
	void draw(){
		fill(getColor());
		noStroke();
		ellipse(position.x, position.y, size, size);
	}
	
	color getColor(){
		return lerpColor(startColor, endColor, constrain(age/lifetime, 0, 1));
	}
}

class ParticleFactory {
	Particle newParticle(){
		return new Particle();
	}
}

class CollisionParticle extends Particle {
	CollisionParticle(){
		startColor = color(255, 110, 2, 250);
		endColor = color(255, 255, 0, 0);
	}
	
	void draw(){
		drawStraightTrail();
		super.draw();
	}
	
	void drawStraightTrail(){
		float trailTime = min(0.45, age);
		Vec2 trailStart = position.sub(velocity.mul(trailTime));
		stroke(getColor());
		line(position.x, position.y, trailStart.x, trailStart.y);
	}
}

class CollisionParticleFactory extends ParticleFactory {
	CollisionParticle newParticle(){
		return new CollisionParticle();
	}
}

class TextParticle extends Particle {
	String content = "[NO TEXT]";
	
	TextParticle(){
		super();
		startColor = color(0, 250);
		endColor = color(0, 1);
		size = 24;
	}
	
	void draw(){
		pushMatrix();
			translate(position.x, position.y);
			scale(1.0 / worldToScreenScale);
			textFont(scoreFont);
			fill(getColor());
			textSize(size);
			text(content, 0, 0);
		popMatrix();
	}
}

class TextParticleFactory extends ParticleFactory {
	TextParticle newParticle(){
		return new TextParticle();
	}
}

class ParticleEmitter extends GameObject {
	protected ArrayList<Particle> particles;
	int limit = 500;
	
	/* Particle properties */
	FloatRange pPosition;      //How much particle position varies from emitter position
	VectorRange pDirection, pAcceleration;
	FloatRange pVelocity, pLifetime, pDamping;
	ParticleFactory factory;
	
	/* Emitter properties */
	Vec2 position;
	float rate;                //particles per second
	float emitterLifetime = 0; //0 = forever
	float emitterAge = 0;	
	PhysicalObject leader = null; //follow this object around
	float elapsedTime = 0;
	
	ParticleEmitter(ParticleFactory factory){
		this.factory = factory;
		particles = new ArrayList<Particle>();
		position = new Vec2(0, 0);
		pPosition = new FloatRange(-1, 1);
		pDirection = new VectorRange(new Vec2(0, -1), radians(360));
		pVelocity = new FloatRange(5.0, 10.0);
		pAcceleration = new VectorRange(new Vec2(0, 0), 0);
		pLifetime = new FloatRange(3, 5);
		pDamping = new FloatRange(0.1, 0.5);
		rate = 20;
	}	
	
	void update(float dt){
		followTheLeader();
		if ( dead ){
			return;
		}
		
		spawnParticles(dt);
		updateParticles(dt);
	}
	
	void followTheLeader(){
		if ( !dead && (leader != null) && (leader.body != null) ){
			if ( leader.dead ){
				die();
				return;
			} else {
				position = leader.getPosition();
			}
		}
	}
	
	void spawnParticles(float dt){
		//Spawn new particles
		elapsedTime += dt;
		float particleInterval = 1.0 / rate;
		while( elapsedTime > particleInterval ){
			emitParticle();
			elapsedTime -= particleInterval;
		}
	}
	
	void updateParticles(float dt){
		//Update particles
		for(int i = particles.size()-1; i >= 0; i--){
			Particle particle = particles.get(i);
			particle.update(dt);
			if ( particle.dead ){
				particles.remove(i);
			}
		}
	}
	
	void draw(){
		for(Particle particle : particles){
			particle.draw();
		}
	}
	
	Particle emitParticle(){
		Particle particle = makeParticle();
		
		if ( particles.size() < limit ){
			particles.add(particle);
		}
		
		return particle;
	}
	
	Particle makeParticle(){
		Particle particle = factory.newParticle();
		
		particle.position = position.add(new Vec2(pPosition.getRandom(), pPosition.getRandom()));
		particle.velocity = pDirection.getRandom().mul(pVelocity.getRandom());
		particle.acceleration = pAcceleration.getRandom();
		particle.lifetime = pLifetime.getRandom();
		particle.damping = pDamping.getRandom();
		
		return particle;
	}
}

class TrailEmitter extends ParticleEmitter {
	Vec2 lastParticlePosition = null;
	
	TrailEmitter(TrailParticleFactory factory){
		super(factory);
		pPosition = new FloatRange(0, 0);
		pDirection = new VectorRange(new Vec2(0, 0), 0);
		pVelocity = new FloatRange(0, 0);
		pAcceleration = new VectorRange(new Vec2(0, 0), 0);
		pLifetime = new FloatRange(3, 5);
		pDamping = new FloatRange(0, 0);
		rate = 20;
	}
	
	void spawnParticles(float dt){
		if ( (leader == null) || (leader.body == null) ){
			return;
		}
		
		if ( lastParticlePosition == null ){
			lastParticlePosition = position.clone();
			return;
		}
		
		Ship ship = (Ship)leader;
		
		TrailParticle particle = (TrailParticle)makeParticle();
		particle.previousPosition = lastParticlePosition;
		particle.position = position;
		particle.size = ship.radius;
		particle.lifetime = 0.2;
		particle.startColor = ship.baseColor;
		particle.endColor = color(255, 0);
		particles.add(particle);
		
		lastParticlePosition = position;
	}
}

class TrailParticle extends Particle {
	Vec2 previousPosition;

	TrailParticle(){
		super();
	}
	
	void draw(){
		fill(getColor());
		noStroke();
		quad(
			position.x, position.y - size,
			position.x, position.y + size,
			previousPosition.x, previousPosition.y + size,
			previousPosition.x, previousPosition.y - size
		);
	}
}

class TrailParticleFactory extends ParticleFactory {
	TrailParticle newParticle(){
		return new TrailParticle();
	}
}

class FloatRange {
	float high, low;
	
	FloatRange(float high, float low){
		this.high = max(low, high);
		this.low = min(low, high);
	}
	
	float getRandom(){
		return random(low, high);
	}
}

class VectorRange {
	Vec2 vector;
	float spread; //in radians
	protected float halfSpread;
	
	VectorRange(Vec2 vector, float spread){
		this.vector = vector;
		this.spread = spread;
		halfSpread = spread/2;
	}
	
	Vec2 getRandom(){
		return rotateVector(vector, random(-halfSpread, halfSpread));
	}
}

class PositionRange {
	Vec2 vector;
	float spread; //radius / half-square
	protected float halfSpread;
	
	PositionRange(Vec2 vector, float spread){
		this.vector = vector;
		this.spread = spread;
		halfSpread = spread/2;
	}
	
	Vec2 getRandom(){
		Vec2 result = vector.clone();
		result.x += random(-halfSpread, halfSpread);
		result.y += random(-halfSpread, halfSpread);
		return result;
	}
}

