class PhysicalObject extends GameObject {
  Body body;
  
  protected Vec2 startPosition; //Where to place the object when it's first added to the Box2D world
  
  void addTo(World world){
    //create the body and add it to the world here
  }
  
  Vec2 getPosition(){
    if ( body != null ){
      return body.getPosition();
    } else {
      return new Vec2(0, 0);
    }
  }
  
  void deInit(){
    if ( body != null ){
      world.destroyBody(body);
      body = null;
    }
  }
  
  void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
    //stub
  }  
  
  void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
    //stub
  }
  
  void onCollisionResult(ContactResult result, PhysicalObject otherObject){
    //stub
  }
}

void initPhysics(){
  //Create the Box2d world
  AABB worldBounds = new AABB(new Vec2(-100f, -100f), new Vec2(100f, 100f));
  boolean doSleep = true;
  world = new World(worldBounds, gravity, doSleep);
  
  //Add walls to the world
  walls = new Body[4];
  
  BodyDef wallBodyDef = new BodyDef();
  PolygonDef wallBox = new PolygonDef();
  wallBox.filter.groupIndex = -1; //walls don't collide with each other
  
  wallBox.setAsBox(wallThickness, height/worldToScreenScale * 0.5);
  wallBodyDef.position.set(-width/worldToScreenScale*0.5 - wallThickness, 0f);
  walls[0] = world.createBody(wallBodyDef);
  walls[0].createShape(wallBox);
  
  wallBodyDef.position.set(width/worldToScreenScale*0.5 + wallThickness, 0f);
  walls[1] = world.createBody(wallBodyDef);
  walls[1].createShape(wallBox);
  
  wallBox.setAsBox(width/worldToScreenScale * 0.5, wallThickness);
  wallBodyDef.position.set(0f, -height/worldToScreenScale*0.5 - wallThickness);
  walls[2] = world.createBody(wallBodyDef);
  walls[2].createShape(wallBox);
  
  wallBodyDef.position.set(0f, height/worldToScreenScale*0.5 + wallThickness);
  walls[3] = world.createBody(wallBodyDef);
  walls[3].createShape(wallBox);

  //Our contact handler/proxy
  contactHandler = new MyContactListener();
  world.setContactListener(contactHandler);
}

void updatePhysics(float dt){
  world.step(dt, 8);
}

class MyContactListener implements ContactListener {
  
  private final int COLLISION_ADD = 1;
  private final int COLLISION_PERSIST = 2;
  private final int COLLISION_REMOVE = 3;
  
  void add(ContactPoint point){
    notifyGameObjects(point, COLLISION_ADD);
    processContact(point);
  }
  
  void persist(ContactPoint point){
    notifyGameObjects(point, COLLISION_PERSIST);
    processContact(point);
  }
  
  void remove(ContactPoint point){
    notifyGameObjects(point, COLLISION_REMOVE);
  }
 
  void result(ContactResult result){
    PhysicalObject obj1 = getObjectFromShape(result.shape1);
    PhysicalObject obj2 = getObjectFromShape(result.shape2);
    
    if ( obj1 != null ){
      obj1.onCollisionResult(result, obj2);
    }
    
    if ( obj2 != null ){
      obj2.onCollisionResult(result, obj1);
    }
	
	spawnCollisionParticles(result);
  }
  
  void notifyGameObjects(ContactPoint point, int phase){
    PhysicalObject obj1 = getObjectFromShape(point.shape1);
    PhysicalObject obj2 = getObjectFromShape(point.shape2);
     
    switch(phase){
      case COLLISION_ADD:
        if ( obj1 != null ){
          obj1.onCollisionStart(point, obj2);
        }
        
        if ( obj2 != null ){
          obj2.onCollisionStart(point, obj1);
        }
        break;
        
      case COLLISION_PERSIST:
        if ( obj1 != null ){
          obj1.onCollisionPersist(point, obj2);
        }
        
        if ( obj2 != null ){
          obj2.onCollisionPersist(point, obj1);
        }
        break;
        
      case COLLISION_REMOVE:
        break;
    }
  }
  
  void spawnCollisionParticles(ContactResult result){
	float impactStrength = calculateImpactStrength(result);
	
	int numParticles = ceil(impactStrength / 10); 
	if ( numParticles < 2 ){
		return;
	}
	
	numParticles = ceil(constrain(numParticles, 12, 64));
	
	float averageSpeed = map(impactStrength, 0, 100, 10, 30);
	final float speedVariance = 0.30;
	FloatRange speed = new FloatRange(averageSpeed * (1 - speedVariance), averageSpeed * (1 + speedVariance));
	
	float averageLifetime = constrain(10 / averageSpeed, 0.2, 0.8);
	final float lifetimeVariance = 0.30;
	FloatRange lifetime = new FloatRange(averageLifetime * (1 - lifetimeVariance), averageLifetime * (1 + lifetimeVariance));
	
	for (int i = 0; i < numParticles; i++ ){
		Particle particle = collisionEmitter.emitParticle();
		particle.position.set(result.position);
		particle.velocity.normalize();
		particle.velocity.mulLocal( speed.getRandom() );
		particle.lifetime = lifetime.getRandom();
	}
  }
  
  void processContact(ContactPoint point){
    Body body1 = point.shape1.getBody();
    Body body2 = point.shape2.getBody();
    
    //Process the hook's collisions with enemy ships
    if ( playerShip.hookActive && (playerShip.hookedBot == null) ){
      Body target = null;
      if ( body1 == playerShip.hook ){
        target = body2; 
      } else if ( body2 == playerShip.hook ){
        target = body1;
      }
      
      //Is the target a bot ship?
      if ( (target != null) && (target.getUserData() != null) && (target.getUserData() instanceof Bot) ){
        Bot targetBot = (Bot)target.getUserData();
        //handle this outside the physics step
        if ( !targetBot.dead ){
          possibleTargets.add( new BotContact(targetBot, point) ); 
        }
      }
    }
  }
}

//Get the game object associated with a Box2d body. 
PhysicalObject getObjectFromBody(Body body){
	if ( body == null ){
	  return null;
	}
	if ( (body.getUserData() != null) && (body.getUserData() instanceof PhysicalObject) ){
	  return (PhysicalObject)body.getUserData();
	}
	return null;    
}

//Get the game object associated with a Box2d shape. If the shape has no associated object, 
//this function will return the game object associated with this shape's body instead.
PhysicalObject getObjectFromShape(Shape shape){
	if ( shape == null ){
	  return null;
	}
	Object obj = shape.getUserData();
	if ( (obj != null) && (obj instanceof PhysicalObject) ){
	  return (PhysicalObject)obj;
	} else {
	  return getObjectFromBody(shape.getBody());
	}
}

//Rotate a 2D vector around the origin. The angle of rotation should be specified in radians.
//Positive values will produce clockwise rotation, negative - counter-clockwise.
Vec2 rotateVector(Vec2 vector, float angle){
  float sin_a = sin(angle), cos_a = cos(angle);
  return new Vec2(
    cos_a * vector.x - sin_a * vector.y,
    sin_a * vector.x + cos_a * vector.y
  );
}

//Calculate the approximate "strength" of a collision. Note: the returned value will likely
//have no particular relation to real collision physics. It's intended to be a relative indication
//of how strong one object hit another, not an accurate measure of absorbed energy or some such.
float calculateImpactStrength(ContactResult result){
	return result.normalImpulse /(result.shape1.getBody().getMass() + result.shape2.getBody().getMass());
}

public abstract class WorldQuery implements Iterator<PhysicalObject>, Iterable<PhysicalObject> {
  private org.jbox2d.dynamics.World world;
  private boolean allowDuplicates = true;
  
  WorldQuery(org.jbox2d.dynamics.World world, boolean allowDuplicates){
    this.world = world;
    this.allowDuplicates = allowDuplicates;
  }
  
  public Iterator<PhysicalObject> iterator(){
    return this;
  }
  
  public boolean hasNext(){
    return false;
  }
  
  public PhysicalObject next(){
    return null;
  }
  
  public void remove(){
    throw new UnsupportedOperationException();
  }
}

/*
 * A helper class for finding all PhysicalObject's that overlap a given AABB
 */
class AABBQuery extends WorldQuery {
  private AABB bounds;
  private List<PhysicalObject> foundEntities;
  private Iterator<PhysicalObject> internalIterator;
    
  AABBQuery(AABB bounds, org.jbox2d.dynamics.World world, boolean allowDuplicates, int maxShapes){
    super(world, allowDuplicates);
    this.bounds = bounds;

    Shape[] foundShapes = world.query(bounds, maxShapes);
    foundEntities = new ArrayList<PhysicalObject>(foundShapes.length);
    for(int i = 0; i < foundShapes.length; i++){
      PhysicalObject data = getObjectFromShape(foundShapes[i]);
      if ( data != null ){
        if ( allowDuplicates || !foundEntities.contains(data) ){
          foundEntities.add((PhysicalObject)data);
        }
      }
    }
    internalIterator = foundEntities.iterator();
  }
  
  AABBQuery(AABB bounds, org.jbox2d.dynamics.World world){
    this(bounds, world, true, 500);
  }
  
  public boolean hasNext(){
    return internalIterator.hasNext();
  }
  
  public PhysicalObject next(){
    return internalIterator.next();
  }
}
