class Pickup extends PhysicalObject {
	float radius = 1;
	float age = 0; //How long the pickup has existed.
	float growthTime = 0.5; //When first spawned, the pickup shall animate to full size within this many seconds
	float lifeTime = 0; //How long the pickup stays around. 0 = no limit.
	float fadeOutTime = 3.0; //How quickly it fades before death
	int bonusPoints = 5; //Get extra points for picking up a powerup
	
	Pickup(Vec2 position, float radius){
		super();
		this.startPosition = position;
		this.radius = radius;
	}
	
	//Check whether a particular entity is allowed to take this pickup
	boolean allowPickupBy(GameObject taker){
		return false;
	}

	//Fired when the pickup is picked up by something
	void onPickup(GameObject taker){
		//implement this in subclasses
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	void draw(){
		Vec2 position = getPosition();
		fill(50, 200, 50);
		stroke(0, 150, 0);
		ellipse(position.x, position.y, getVisibleRadius()*2, getVisibleRadius()*2);
	}
	
	float getVisibleRadius(){
		if ( age >= growthTime ){
			return radius;
		} else {
			return age/growthTime*radius;
		}
	}
	
	int getAlpha(){
		if (lifeTime == 0){
			return 255;
		}
		return round(map( 
			constrain((lifeTime - age)/fadeOutTime, 0, 1),
			0, 1, 0, 255
		));
	}
	
	void update(float dt){
		age += dt;
		if ( (lifeTime != 0) && (age > lifeTime) ){
			die();
		}
	}
	
	void addTo(World world){
		//Create the Box2d body corresponding to this pickup and add it to the physical world
		BodyDef pickupBodyDef = new BodyDef();
		pickupBodyDef.position.x = startPosition.x;
		pickupBodyDef.position.y = startPosition.y;
		pickupBodyDef.fixedRotation = true;
		
		CircleDef pickupShape = new CircleDef();
		pickupShape.radius = radius;
		pickupShape.density = 0;
		pickupShape.isSensor = true; //pass through other bodies harmlessly
		
		body = world.createBody(pickupBodyDef);
		body.createShape(pickupShape);
		body.setMassFromShapes();
		body.setUserData(this);
	}
	
	protected void handleCollision(ContactPoint contact, PhysicalObject otherObject){
		if ( dead || (otherObject == null) ){
			return;
		}
		if (allowPickupBy(otherObject)){
			onPickup(otherObject);
			if ( otherObject == playerShip ){
				//Note: bad design. What if something else picks up the powerup?
				awardPoints(bonusPoints, 0, getPosition());
			}
			die();
		}
	}
	
	void onCollisionStart(ContactPoint contact, PhysicalObject otherObject){
		handleCollision(contact, otherObject);
	}  
	  
	void onCollisionPersist(ContactPoint contact, PhysicalObject otherObject){
		handleCollision(contact, otherObject);
	}
}

class HealthPickup extends Pickup {
	float healthIncrease = 50;
	
	HealthPickup(Vec2 position, float radius){
		super(position, radius);
		lifeTime = 20;
	}
	
	HealthPickup(Vec2 position){
		this(position, 2.0);
	}

	//The player ship can  pick up health when it's damaged but not yet dead
	boolean allowPickupBy(GameObject taker){
		if ( !(taker instanceof PlayerShip) || taker.dead){
			return false;
		};
		PlayerShip player = (PlayerShip)taker;
		return (player.hitpoints < player.maxHitpoints);
	}

	//Heal the player
	void onPickup(GameObject taker){
		PlayerShip player = (PlayerShip)taker;
		player.heal(healthIncrease);
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	void draw(){
		Vec2 position = getPosition();
		pushMatrix();		
			translate(position.x, position.y);
			float vradius = getVisibleRadius();
			int valpha = getAlpha();
			//The green base circle
			fill(50, 200, 50, valpha);
			stroke(0, 150, 0, valpha);
			ellipse(0, 0, vradius*2, vradius*2);
			//The cross
			rectMode(CENTER);
			fill(255, valpha);
			noStroke();
			float cw = vradius*1.3, ch = vradius/2.5;
			rect(0, 0, cw, ch);
			rect(0, 0, ch, cw);
		popMatrix();
	}
}

class PowerPickup extends Pickup {
	float powerIncrease = 10;
	
	PowerPickup(Vec2 position, float radius){
		super(position, radius);
		lifeTime = 10;
	}
	
	PowerPickup(Vec2 position){
		this(position, 1.0);
	}

	//Only the player can pick up power
	boolean allowPickupBy(GameObject taker){
		return (taker instanceof PlayerShip) && (!taker.dead);
	}

	//Give the player more power and +1 HP
	void onPickup(GameObject taker){
		PlayerShip player = (PlayerShip)taker;
		player.power = constrain(player.power + powerIncrease, 0, player.maxPower);
		player.heal(1);
	}
	
	//Draw the pickup. Should be overriden in a subclass.
	void draw(){
		Vec2 position = getPosition();
		pushMatrix();		
			translate(position.x, position.y);
			float vradius = getVisibleRadius();
			int valpha = getAlpha();
			//The base circle
			fill(50, 200, 50, valpha);
			stroke(0, 150, 0, valpha);
			ellipse(0, 0, vradius*2, vradius*2);
			//The triangle
			fill(255, valpha);
			noStroke();
			triangle(
				0, -vradius * 0.7,
				-vradius * 0.6, vradius * 0.6,
				vradius * 0.6, vradius * 0.6
			);
		popMatrix();
	}
}
